# Licensed under a 3-clause BSD style license - see LICENSE.rst

from collections import OrderedDict, UserDict
from collections.abc import Mapping

import numpy as np
import pytest

import astropy.units as u
from astropy.table import Column, MaskedColumn, QTable, Table, TableColumns


class DictLike(Mapping):
    """A minimal mapping-like object that does not subclass dict.

    This is used to test code that expects dict-like but without actually
    inheriting from dict.
    """

    def __init__(self, *args, **kwargs):
        self._data = dict(*args, **kwargs)

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, item, value):
        self._data[item] = value

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)


class TestTableColumnsInit:
    def test_init(self):
        """Test initialisation with lists, tuples, dicts of arrays
        rather than Columns [regression test for #2647]"""
        x1 = np.arange(10.0)
        x2 = np.arange(5.0)
        x3 = np.arange(7.0)
        col_list = [("x1", x1), ("x2", x2), ("x3", x3)]
        tc_list = TableColumns(col_list)
        for col in col_list:
            assert col[0] in tc_list
            assert tc_list[col[0]] is col[1]

        col_tuple = (("x1", x1), ("x2", x2), ("x3", x3))
        tc_tuple = TableColumns(col_tuple)
        for col in col_tuple:
            assert col[0] in tc_tuple
            assert tc_tuple[col[0]] is col[1]

        col_dict = dict([("x1", x1), ("x2", x2), ("x3", x3)])
        tc_dict = TableColumns(col_dict)
        for col in tc_dict.keys():
            assert col in tc_dict
            assert tc_dict[col] is col_dict[col]

        columns = [Column(col[1], name=col[0]) for col in col_list]
        tc = TableColumns(columns)
        for col in columns:
            assert col.name in tc
            assert tc[col.name] is col


# pytest.mark.usefixtures('table_type')
class BaseInitFrom:
    def _setup(self, table_type):
        pass

    def test_basic_init(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, names=("a", "b", "c"))
        assert t.colnames == ["a", "b", "c"]
        assert np.all(t["a"] == np.array([1, 3]))
        assert np.all(t["b"] == np.array([2, 4]))
        assert np.all(t["c"] == np.array([3, 5]))
        assert all(t[name].name == name for name in t.colnames)

    def test_set_dtype(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, names=("a", "b", "c"), dtype=("i4", "f4", "f8"))
        assert t.colnames == ["a", "b", "c"]
        assert np.all(t["a"] == np.array([1, 3], dtype="i4"))
        assert np.all(t["b"] == np.array([2, 4], dtype="f4"))
        assert np.all(t["c"] == np.array([3, 5], dtype="f8"))
        assert t["a"].dtype.type == np.int32
        assert t["b"].dtype.type == np.float32
        assert t["c"].dtype.type == np.float64
        assert all(t[name].name == name for name in t.colnames)

    def test_names_dtype_mismatch(self, table_type):
        self._setup(table_type)
        with pytest.raises(ValueError):
            table_type(self.data, names=("a",), dtype=("i4", "f4", "i4"))

    def test_names_cols_mismatch(self, table_type):
        self._setup(table_type)
        with pytest.raises(ValueError):
            table_type(self.data, names=("a",), dtype="i4")


@pytest.mark.usefixtures("table_type")
class BaseInitFromListLike(BaseInitFrom):
    def test_names_cols_mismatch(self, table_type):
        self._setup(table_type)
        with pytest.raises(ValueError):
            table_type(self.data, names=["a"], dtype=[int])

    def test_names_copy_false(self, table_type):
        self._setup(table_type)
        with pytest.raises(ValueError):
            table_type(self.data, names=["a"], dtype=[int], copy=False)


@pytest.mark.usefixtures("table_type")
class BaseInitFromDictLike(BaseInitFrom):
    pass


@pytest.mark.usefixtures("table_type")
class TestInitFromNdarrayHomo(BaseInitFromListLike):
    def setup_method(self, method):
        self.data = np.array([(1, 2, 3), (3, 4, 5)], dtype="i4")

    def test_default_names(self, table_type):
        self._setup(table_type)
        t = table_type(self.data)
        assert t.colnames == ["col0", "col1", "col2"]

    def test_ndarray_ref(self, table_type):
        """Init with ndarray and copy=False and show that this is a reference
        to input ndarray"""
        self._setup(table_type)
        t = table_type(self.data, copy=False)
        t["col1"][1] = 0
        assert t.as_array()["col1"][1] == 0
        assert t["col1"][1] == 0
        assert self.data[1][1] == 0

    def test_partial_names_dtype(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, names=["a", None, "c"], dtype=[None, None, "f8"])
        assert t.colnames == ["a", "col1", "c"]
        assert t["a"].dtype.type == np.int32
        assert t["col1"].dtype.type == np.int32
        assert t["c"].dtype.type == np.float64
        assert all(t[name].name == name for name in t.colnames)

    def test_partial_names_ref(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, names=["a", None, "c"])
        assert t.colnames == ["a", "col1", "c"]
        assert t["a"].dtype.type == np.int32
        assert t["col1"].dtype.type == np.int32
        assert t["c"].dtype.type == np.int32
        assert all(t[name].name == name for name in t.colnames)


@pytest.mark.usefixtures("table_type")
class TestInitFromListOfLists(BaseInitFromListLike):
    def setup_method(self, table_type):
        self._setup(table_type)
        self.data = [
            (np.int32(1), np.int32(3)),
            Column(name="col1", data=[2, 4], dtype=np.int32),
            np.array([3, 5], dtype=np.int32),
        ]

    def test_default_names(self, table_type):
        self._setup(table_type)
        t = table_type(self.data)
        assert t.colnames == ["col0", "col1", "col2"]
        assert all(t[name].name == name for name in t.colnames)

    def test_partial_names_dtype(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, names=["b", None, "c"], dtype=["f4", None, "f8"])
        assert t.colnames == ["b", "col1", "c"]
        assert t["b"].dtype.type == np.float32
        assert t["col1"].dtype.type == np.int32
        assert t["c"].dtype.type == np.float64
        assert all(t[name].name == name for name in t.colnames)

    def test_bad_data(self, table_type):
        self._setup(table_type)
        with pytest.raises(ValueError):
            table_type([[1, 2], [3, 4, 5]])


@pytest.mark.usefixtures("table_type")
class TestInitFromListOfDicts(BaseInitFromListLike):
    def _setup(self, table_type):
        self.data = [{"a": 1, "b": 2, "c": 3}, {"a": 3, "b": 4, "c": 5}]
        self.data_ragged = [{"a": 1, "b": 2}, {"a": 2, "c": 4}]

    def test_names(self, table_type):
        self._setup(table_type)
        t = table_type(self.data)
        assert all(colname in {"a", "b", "c"} for colname in t.colnames)

    def test_names_ordered(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, names=("c", "b", "a"))
        assert t.colnames == ["c", "b", "a"]

    def test_missing_data_init_from_dict(self, table_type):
        self._setup(table_type)
        dat = self.data_ragged
        for rows in [False, True]:
            t = table_type(rows=dat) if rows else table_type(dat)

            assert np.all(t["a"] == [1, 2])
            assert np.all(t["b"].mask == [False, True])
            assert np.all(t["b"].data == [2, 2])
            assert np.all(t["c"].mask == [True, False])
            assert np.all(t["c"].data == [4, 4])

            assert type(t["a"]) is (MaskedColumn if t.masked else Column)
            assert type(t["b"]) is MaskedColumn
            assert type(t["c"]) is MaskedColumn


class TestInitFromListOfMapping(TestInitFromListOfDicts):
    """Test that init from a Mapping that is not a dict subclass works"""

    def _setup(self, table_type):
        self.data = [DictLike(a=1, b=2, c=3), DictLike(a=3, b=4, c=5)]
        self.data_ragged = [DictLike(a=1, b=2), DictLike(a=2, c=4)]
        # Make sure data rows are not a dict subclass
        assert not isinstance(self.data[0], dict)


@pytest.mark.usefixtures("table_type")
class TestInitFromColsList(BaseInitFromListLike):
    def _setup(self, table_type):
        self.data = [
            Column([1, 3], name="x", dtype=np.int32),
            np.array([2, 4], dtype=np.int32),
            np.array([3, 5], dtype="i8"),
        ]

    def test_default_names(self, table_type):
        self._setup(table_type)
        t = table_type(self.data)
        assert t.colnames == ["x", "col1", "col2"]
        assert all(t[name].name == name for name in t.colnames)

    def test_partial_names_dtype(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, names=["b", None, "c"], dtype=["f4", None, "f8"])
        assert t.colnames == ["b", "col1", "c"]
        assert t["b"].dtype.type == np.float32
        assert t["col1"].dtype.type == np.int32
        assert t["c"].dtype.type == np.float64
        assert all(t[name].name == name for name in t.colnames)

    def test_ref(self, table_type):
        """Test that initializing from a list of columns can be done by reference"""
        self._setup(table_type)
        t = table_type(self.data, copy=False)
        t["x"][0] = 100
        assert self.data[0][0] == 100


@pytest.mark.usefixtures("table_type")
class TestInitFromNdarrayStruct(BaseInitFromDictLike):
    def _setup(self, table_type):
        self.data = np.array(
            [(1, 2, 3), (3, 4, 5)], dtype=[("x", "i8"), ("y", "i4"), ("z", "i8")]
        )

    def test_ndarray_ref(self, table_type):
        """Init with ndarray and copy=False and show that table uses reference
        to input ndarray"""
        self._setup(table_type)
        t = table_type(self.data, copy=False)

        t["x"][1] = 0  # Column-wise assignment
        t[0]["y"] = 0  # Row-wise assignment
        assert self.data["x"][1] == 0
        assert self.data["y"][0] == 0
        assert np.all(np.array(t) == self.data)
        assert all(t[name].name == name for name in t.colnames)

    def test_partial_names_dtype(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, names=["e", None, "d"], dtype=["f4", None, "f8"])
        assert t.colnames == ["e", "y", "d"]
        assert t["e"].dtype.type == np.float32
        assert t["y"].dtype.type == np.int32
        assert t["d"].dtype.type == np.float64
        assert all(t[name].name == name for name in t.colnames)

    def test_partial_names_ref(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, names=["e", None, "d"], copy=False)
        assert t.colnames == ["e", "y", "d"]
        assert t["e"].dtype.type == np.int64
        assert t["y"].dtype.type == np.int32
        assert t["d"].dtype.type == np.int64
        assert all(t[name].name == name for name in t.colnames)


@pytest.mark.usefixtures("table_type")
class TestInitFromDict(BaseInitFromDictLike):
    def _setup(self, table_type):
        self.data = dict(
            [
                ("a", Column([1, 3], name="x")),
                ("b", [2, 4]),
                ("c", np.array([3, 5], dtype="i8")),
            ]
        )


@pytest.mark.usefixtures("table_type")
class TestInitFromMapping(BaseInitFromDictLike):
    def _setup(self, table_type):
        self.data = UserDict(
            [
                ("a", Column([1, 3], name="x")),
                ("b", [2, 4]),
                ("c", np.array([3, 5], dtype="i8")),
            ]
        )
        assert isinstance(self.data, Mapping)
        assert not isinstance(self.data, dict)


@pytest.mark.usefixtures("table_type")
class TestInitFromOrderedDict(BaseInitFromDictLike):
    def _setup(self, table_type):
        self.data = OrderedDict(
            [
                ("a", Column(name="x", data=[1, 3])),
                ("b", [2, 4]),
                ("c", np.array([3, 5], dtype="i8")),
            ]
        )

    def test_col_order(self, table_type):
        self._setup(table_type)
        t = table_type(self.data)
        assert t.colnames == ["a", "b", "c"]


@pytest.mark.usefixtures("table_type")
class TestInitFromRow(BaseInitFromDictLike):
    def _setup(self, table_type):
        arr = np.array(
            [(1, 2, 3), (3, 4, 5)], dtype=[("x", "i8"), ("y", "i8"), ("z", "f8")]
        )
        self.data = table_type(arr, meta={"comments": ["comment1", "comment2"]})

    def test_init_from_row(self, table_type):
        self._setup(table_type)
        t = table_type(self.data[0])

        # Values and meta match original
        assert t.meta["comments"][0] == "comment1"
        for name in t.colnames:
            assert np.all(t[name] == self.data[name][0:1])
        assert all(t[name].name == name for name in t.colnames)

        # Change value in new instance and check that original is the same
        t["x"][0] = 8
        t.meta["comments"][1] = "new comment2"
        assert np.all(t["x"] == np.array([8]))
        assert np.all(self.data["x"] == np.array([1, 3]))
        assert self.data.meta["comments"][1] == "comment2"


@pytest.mark.usefixtures("table_type")
class TestInitFromTable(BaseInitFromDictLike):
    def _setup(self, table_type):
        arr = np.array(
            [(1, 2, 3), (3, 4, 5)], dtype=[("x", "i8"), ("y", "i8"), ("z", "f8")]
        )
        self.data = table_type(arr, meta={"comments": ["comment1", "comment2"]})

    def test_data_meta_copy(self, table_type):
        self._setup(table_type)
        t = table_type(self.data)
        assert t.meta["comments"][0] == "comment1"
        t["x"][1] = 8
        t.meta["comments"][1] = "new comment2"
        assert self.data.meta["comments"][1] == "comment2"
        assert np.all(t["x"] == np.array([1, 8]))
        assert np.all(self.data["x"] == np.array([1, 3]))
        assert t["z"].name == "z"
        assert all(t[name].name == name for name in t.colnames)

    def test_table_ref(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, copy=False)
        t["x"][1] = 0
        assert t["x"][1] == 0
        assert self.data["x"][1] == 0
        assert np.all(t.as_array() == self.data.as_array())
        assert all(t[name].name == name for name in t.colnames)

    def test_partial_names_dtype(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, names=["e", None, "d"], dtype=["f4", None, "i8"])
        assert t.colnames == ["e", "y", "d"]
        assert t["e"].dtype.type == np.float32
        assert t["y"].dtype.type == np.int64
        assert t["d"].dtype.type == np.int64
        assert all(t[name].name == name for name in t.colnames)

    def test_partial_names_ref(self, table_type):
        self._setup(table_type)
        t = table_type(self.data, names=["e", None, "d"], copy=False)
        assert t.colnames == ["e", "y", "d"]
        assert t["e"].dtype.type == np.int64
        assert t["y"].dtype.type == np.int64
        assert t["d"].dtype.type == np.float64
        assert all(t[name].name == name for name in t.colnames)

    def test_init_from_columns(self, table_type):
        self._setup(table_type)
        t = table_type(self.data)
        t2 = table_type(t.columns["z", "x", "y"])
        assert t2.colnames == ["z", "x", "y"]
        assert t2.dtype.names == ("z", "x", "y")

    def test_init_from_columns_slice(self, table_type):
        self._setup(table_type)
        t = table_type(self.data)
        t2 = table_type(t.columns[0:2])
        assert t2.colnames == ["x", "y"]
        assert t2.dtype.names == ("x", "y")

    def test_init_from_columns_mix(self, table_type):
        self._setup(table_type)
        t = table_type(self.data)
        t2 = table_type([t.columns[0], t.columns["z"]])
        assert t2.colnames == ["x", "z"]
        assert t2.dtype.names == ("x", "z")


@pytest.mark.usefixtures("table_type")
class TestInitFromNone:
    # Note table_table.TestEmptyData tests initializing a completely empty
    # table and adding data.

    def test_data_none_with_cols(self, table_type):
        """
        Test different ways of initing an empty table
        """
        np_t = np.empty(0, dtype=[("a", "f4", (2,)), ("b", "i4")])
        for kwargs in (
            {"names": ("a", "b")},
            {"names": ("a", "b"), "dtype": (("f4", (2,)), "i4")},
            {"dtype": [("a", "f4", (2,)), ("b", "i4")]},
            {"dtype": np_t.dtype},
        ):
            t = table_type(**kwargs)
            assert t.colnames == ["a", "b"]
            assert len(t["a"]) == 0
            assert len(t["b"]) == 0
            if "dtype" in kwargs:
                assert t["a"].dtype.type == np.float32
                assert t["b"].dtype.type == np.int32
                assert t["a"].shape[1:] == (2,)


@pytest.mark.usefixtures("table_types")
class TestInitFromRows:
    def test_init_with_rows(self, table_type):
        for rows in ([[1, "a"], [2, "b"]], [(1, "a"), (2, "b")], ((1, "a"), (2, "b"))):
            t = table_type(rows=rows, names=("a", "b"))
            assert np.all(t["a"] == [1, 2])
            assert np.all(t["b"] == ["a", "b"])
            assert t.colnames == ["a", "b"]
            assert t["a"].dtype.kind == "i"
            assert t["b"].dtype.kind in ("S", "U")
            # Regression test for
            # https://github.com/astropy/astropy/issues/3052
            assert t["b"].dtype.str.endswith("1")

        rows = np.arange(6).reshape(2, 3)
        t = table_type(rows=rows, names=("a", "b", "c"), dtype=["f8", "f4", "i8"])
        assert np.all(t["a"] == [0, 3])
        assert np.all(t["b"] == [1, 4])
        assert np.all(t["c"] == [2, 5])
        assert t.colnames == ["a", "b", "c"]
        assert t["a"].dtype.str.endswith("f8")
        assert t["b"].dtype.str.endswith("f4")
        assert t["c"].dtype.str.endswith("i8")

    def test_init_with_rows_and_data(self, table_type):
        with pytest.raises(ValueError) as err:
            table_type(data=[[1]], rows=[[1]])
        assert "Cannot supply both `data` and `rows` values" in str(err.value)


@pytest.mark.parametrize("has_data", [True, False])
def test_init_table_with_names_and_structured_dtype(has_data):
    """Test fix for #10393"""
    arr = np.ones(2, dtype=np.dtype([("a", "i4"), ("b", "f4")]))
    data_args = [arr] if has_data else []
    t = Table(*data_args, names=["x", "y"], dtype=arr.dtype)
    assert t.colnames == ["x", "y"]
    assert str(t["x"].dtype) == "int32"
    assert str(t["y"].dtype) == "float32"
    assert len(t) == (2 if has_data else 0)


@pytest.mark.usefixtures("table_type")
def test_init_and_ref_from_multidim_ndarray(table_type):
    """
    Test that initializing from an ndarray structured array with
    a multi-dim column works for both copy=False and True and that
    the referencing is as expected.
    """
    for copy in (False, True):
        nd = np.array(
            [(1, [10, 20]), (3, [30, 40])], dtype=[("a", "i8"), ("b", "i8", (2,))]
        )
        t = table_type(nd, copy=copy)
        assert t.colnames == ["a", "b"]
        assert t["a"].shape == (2,)
        assert t["b"].shape == (2, 2)
        t["a"][0] = -200
        t["b"][1][1] = -100
        if copy:
            assert nd["a"][0] == 1
            assert nd["b"][1][1] == 40
        else:
            assert nd["a"][0] == -200
            assert nd["b"][1][1] == -100


@pytest.mark.usefixtures("table_type")
@pytest.mark.parametrize("copy", [False, True])
def test_init_and_ref_from_dict(table_type, copy):
    """
    Test that initializing from a dict works for both copy=False and True and that
    the referencing is as expected.
    """
    x1 = np.arange(10.0)
    x2 = np.zeros(10)
    col_dict = dict([("x1", x1), ("x2", x2)])
    t = table_type(col_dict, copy=copy)
    assert set(t.colnames) == {"x1", "x2"}
    assert t["x1"].shape == (10,)
    assert t["x2"].shape == (10,)
    t["x1"][0] = -200
    t["x2"][1] = -100
    if copy:
        assert x1[0] == 0.0
        assert x2[1] == 0.0
    else:
        assert x1[0] == -200
        assert x2[1] == -100


def test_add_none_object_column():
    """Test fix for a problem introduced in #10636 (see
    https://github.com/astropy/astropy/pull/10636#issuecomment-676847515)
    """
    t = Table(data={"a": [1, 2, 3]})
    t["b"] = None
    assert all(val is None for val in t["b"])
    assert t["b"].dtype.kind == "O"


@pytest.mark.usefixtures("table_type")
def test_init_from_row_OrderedDict(table_type):
    row1 = OrderedDict([("b", 1), ("a", 0)])
    row2 = {"a": 10, "b": 20}
    rows12 = [row1, row2]
    row3 = dict([("b", 1), ("a", 0)])
    row4 = dict([("b", 11), ("a", 10)])
    rows34 = [row3, row4]
    t1 = table_type(rows=rows12)
    t2 = table_type(rows=rows34)
    t3 = t2[sorted(t2.colnames)]
    assert t1.colnames == ["b", "a"]
    assert t2.colnames == ["b", "a"]
    assert t3.colnames == ["a", "b"]


def test_init_from_rows_as_generator():
    rows = ((1 + ii, 2 + ii) for ii in range(2))
    t = Table(rows=rows)
    assert np.all(t["col0"] == [1, 2])
    assert np.all(t["col1"] == [2, 3])


@pytest.mark.parametrize("dtype", ["fail", "i4"])
def test_init_bad_dtype_in_empty_table(dtype):
    with pytest.raises(
        ValueError, match="type was specified but could not be parsed for column names"
    ):
        Table(dtype=dtype)


def test_init_data_type_not_allowed_to_init_table():
    with pytest.raises(
        ValueError, match="Data type <class 'str'> not allowed to init Table"
    ):
        Table("hello")


def test_init_Table_from_list_of_quantity():
    """Test fix for #11327"""
    # Variation on original example in #11327 at the Table level
    data = [{"x": 5 * u.m, "y": 1 * u.m}, {"x": 10 * u.m, "y": 3}]
    t = Table(data)
    assert t["x"].unit is u.m
    assert t["y"].unit is None
    assert t["x"].dtype.kind == "f"
    assert t["y"].dtype.kind == "O"
    assert np.all(t["x"] == [5, 10])
    assert t["y"][0] == 1 * u.m
    assert t["y"][1] == 3


def test_init_QTable_and_set_units():
    """
    Test fix for #14336 where providing units to QTable init fails.

    This applies when the input is a Quantity.
    """
    t = QTable([[1, 2] * u.km, [1, 2]], units={"col0": u.m, "col1": u.s})
    assert t["col0"].unit == u.m
    assert np.all(t["col0"].value == [1000, 2000])
    assert t["col1"].unit == u.s
    assert np.all(t["col1"].value == [1, 2])



import numpy as np
import pytest
from matplotlib.lines import Path

from astropy.visualization.wcsaxes.grid_paths import get_lon_lat_path


@pytest.mark.parametrize("step_in_degrees", [10, 1, 0.01])
def test_round_trip_visibility(step_in_degrees):
    zero = np.zeros(100)

    # The pixel values are irrelevant for this test
    pixel = np.stack([zero, zero]).T

    # Create a grid line of constant latitude with a point every step
    line = np.stack([np.arange(100), zero]).T * step_in_degrees

    # Create a modified grid line where the point spacing is larger by 5%
    # Starting with point 20, the discrepancy between `line` and `line_round` is greater than `step`
    line_round = line * 1.05

    # Perform the round-trip check
    path = get_lon_lat_path(line, pixel, line_round)

    # The grid line should be visible for only the initial part line (19 points)
    codes_check = np.full(100, Path.MOVETO)
    codes_check[line_round[:, 0] - line[:, 0] < step_in_degrees] = Path.LINETO
    assert np.all(path.codes[1:] == codes_check[1:])



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import itertools
from contextlib import nullcontext

import numpy as np
import numpy.ma as ma
import pytest
from numpy.testing import (
    assert_allclose,
    assert_array_almost_equal,
    assert_array_almost_equal_nulp,
)

from astropy import units as u
from astropy.convolution.convolve import convolve, convolve_fft
from astropy.convolution.kernels import Gaussian2DKernel
from astropy.utils.compat.optional_deps import HAS_PANDAS, HAS_SCIPY
from astropy.utils.exceptions import AstropyUserWarning

VALID_DTYPES = (">f4", "<f4", ">f8", "<f8")
VALID_DTYPE_MATRIX = list(itertools.product(VALID_DTYPES, VALID_DTYPES))

BOUNDARY_OPTIONS = [None, "fill", "wrap", "extend"]
NANHANDLING_OPTIONS = ["interpolate", "fill"]
NORMALIZE_OPTIONS = [True, False]
PRESERVE_NAN_OPTIONS = [True, False]

BOUNDARIES_AND_CONVOLUTIONS = list(
    zip(itertools.cycle((convolve,)), BOUNDARY_OPTIONS)
) + [(convolve_fft, "wrap"), (convolve_fft, "fill")]


class TestConvolve1D:
    def test_list(self):
        """
        Test that convolve works correctly when inputs are lists
        """

        x = [1, 4, 5, 6, 5, 7, 8]
        y = [0.2, 0.6, 0.2]
        z = convolve(x, y, boundary=None)
        assert_array_almost_equal_nulp(
            z, np.array([0.0, 3.6, 5.0, 5.6, 5.6, 6.8, 0.0]), 10
        )

    def test_tuple(self):
        """
        Test that convolve works correctly when inputs are tuples
        """

        x = (1, 4, 5, 6, 5, 7, 8)
        y = (0.2, 0.6, 0.2)
        z = convolve(x, y, boundary=None)
        assert_array_almost_equal_nulp(
            z, np.array([0.0, 3.6, 5.0, 5.6, 5.6, 6.8, 0.0]), 10
        )

    @pytest.mark.parametrize(
        ("boundary", "nan_treatment", "normalize_kernel", "preserve_nan", "dtype"),
        itertools.product(
            BOUNDARY_OPTIONS,
            NANHANDLING_OPTIONS,
            NORMALIZE_OPTIONS,
            PRESERVE_NAN_OPTIONS,
            VALID_DTYPES,
        ),
    )
    def test_quantity(
        self, boundary, nan_treatment, normalize_kernel, preserve_nan, dtype
    ):
        """
        Test that convolve works correctly when input array is a Quantity
        """

        x = np.array([1, 4, 5, 6, 5, 7, 8], dtype=dtype) * u.ph
        y = np.array([0.2, 0.6, 0.2], dtype=dtype)
        z = convolve(
            x,
            y,
            boundary=boundary,
            nan_treatment=nan_treatment,
            normalize_kernel=normalize_kernel,
            preserve_nan=preserve_nan,
        )

        assert x.unit == z.unit

    @pytest.mark.parametrize(
        ("boundary", "nan_treatment", "normalize_kernel", "preserve_nan", "dtype"),
        itertools.product(
            BOUNDARY_OPTIONS,
            NANHANDLING_OPTIONS,
            NORMALIZE_OPTIONS,
            PRESERVE_NAN_OPTIONS,
            VALID_DTYPES,
        ),
    )
    def test_input_unmodified(
        self, boundary, nan_treatment, normalize_kernel, preserve_nan, dtype
    ):
        """
        Test that convolve works correctly when inputs are lists
        """

        array = [1.0, 4.0, 5.0, 6.0, 5.0, 7.0, 8.0]
        kernel = [0.2, 0.6, 0.2]
        x = np.array(array, dtype=dtype)
        y = np.array(kernel, dtype=dtype)

        # Make pseudoimmutable
        x.flags.writeable = False
        y.flags.writeable = False

        convolve(
            x,
            y,
            boundary=boundary,
            nan_treatment=nan_treatment,
            normalize_kernel=normalize_kernel,
            preserve_nan=preserve_nan,
        )

        assert np.all(np.array(array, dtype=dtype) == x)
        assert np.all(np.array(kernel, dtype=dtype) == y)

    @pytest.mark.parametrize(
        ("boundary", "nan_treatment", "normalize_kernel", "preserve_nan", "dtype"),
        itertools.product(
            BOUNDARY_OPTIONS,
            NANHANDLING_OPTIONS,
            NORMALIZE_OPTIONS,
            PRESERVE_NAN_OPTIONS,
            VALID_DTYPES,
        ),
    )
    def test_input_unmodified_with_nan(
        self, boundary, nan_treatment, normalize_kernel, preserve_nan, dtype
    ):
        """
        Test that convolve doesn't modify the input data
        """

        array = [1.0, 4.0, 5.0, np.nan, 5.0, 7.0, 8.0]
        kernel = [0.2, 0.6, 0.2]
        x = np.array(array, dtype=dtype)
        y = np.array(kernel, dtype=dtype)

        # Make pseudoimmutable
        x.flags.writeable = False
        y.flags.writeable = False

        # make copies for post call comparison
        x_copy = x.copy()
        y_copy = y.copy()

        convolve(
            x,
            y,
            boundary=boundary,
            nan_treatment=nan_treatment,
            normalize_kernel=normalize_kernel,
            preserve_nan=preserve_nan,
        )

        # ( NaN == NaN ) = False
        # Only compare non NaN values for canonical equivalence
        # and then check NaN explicitly with np.isnan()
        array_is_nan = np.isnan(array)
        kernel_is_nan = np.isnan(kernel)
        array_not_nan = ~array_is_nan
        kernel_not_nan = ~kernel_is_nan
        assert np.all(x_copy[array_not_nan] == x[array_not_nan])
        assert np.all(y_copy[kernel_not_nan] == y[kernel_not_nan])
        assert np.all(np.isnan(x[array_is_nan]))
        assert np.all(np.isnan(y[kernel_is_nan]))

    @pytest.mark.parametrize(("dtype_array", "dtype_kernel"), VALID_DTYPE_MATRIX)
    def test_dtype(self, dtype_array, dtype_kernel):
        """
        Test that 32- and 64-bit floats are correctly handled
        """

        x = np.array([1.0, 2.0, 3.0], dtype=dtype_array)

        y = np.array([0.0, 1.0, 0.0], dtype=dtype_kernel)

        z = convolve(x, y)

        assert x.dtype == z.dtype

    @pytest.mark.parametrize(
        (
            "convfunc",
            "boundary",
        ),
        BOUNDARIES_AND_CONVOLUTIONS,
    )
    def test_unity_1_none(self, boundary, convfunc):
        """
        Test that a unit kernel with a single element returns the same array
        """

        x = np.array([1.0, 2.0, 3.0], dtype=">f8")

        y = np.array([1.0], dtype=">f8")

        z = convfunc(x, y, boundary=boundary)

        np.testing.assert_allclose(z, x)

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_unity_3(self, boundary):
        """
        Test that a unit kernel with three elements returns the same array
        (except when boundary is None).
        """

        x = np.array([1.0, 2.0, 3.0], dtype=">f8")

        y = np.array([0.0, 1.0, 0.0], dtype=">f8")

        z = convolve(x, y, boundary=boundary)

        if boundary is None:
            assert np.all(z == np.array([0.0, 2.0, 0.0], dtype=">f8"))
        else:
            assert np.all(z == x)

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_uniform_3(self, boundary):
        """
        Test that the different modes are producing the correct results using
        a uniform kernel with three elements
        """

        x = np.array([1.0, 0.0, 3.0], dtype=">f8")

        y = np.array([1.0, 1.0, 1.0], dtype=">f8")

        z = convolve(x, y, boundary=boundary, normalize_kernel=False)

        if boundary is None:
            assert np.all(z == np.array([0.0, 4.0, 0.0], dtype=">f8"))
        elif boundary == "fill":
            assert np.all(z == np.array([1.0, 4.0, 3.0], dtype=">f8"))
        elif boundary == "wrap":
            assert np.all(z == np.array([4.0, 4.0, 4.0], dtype=">f8"))
        else:
            assert np.all(z == np.array([2.0, 4.0, 6.0], dtype=">f8"))

    @pytest.mark.parametrize(
        ("boundary", "nan_treatment", "normalize_kernel", "preserve_nan"),
        itertools.product(
            BOUNDARY_OPTIONS,
            NANHANDLING_OPTIONS,
            NORMALIZE_OPTIONS,
            PRESERVE_NAN_OPTIONS,
        ),
    )
    def test_unity_3_withnan(
        self, boundary, nan_treatment, normalize_kernel, preserve_nan
    ):
        """
        Test that a unit kernel with three elements returns the same array
        (except when boundary is None). This version includes a NaN value in
        the original array.
        """

        x = np.array([1.0, np.nan, 3.0], dtype=">f8")

        y = np.array([0.0, 1.0, 0.0], dtype=">f8")

        # Since the kernel is actually only one pixel wide (because of the
        # zeros) the NaN value doesn't get interpolated over so a warning is
        # expected.
        if nan_treatment == "interpolate" and not preserve_nan:
            ctx = pytest.warns(
                AstropyUserWarning,
                match="nan_treatment='interpolate', however, NaN values detected",
            )
        else:
            ctx = nullcontext()

        with ctx:
            z = convolve(
                x,
                y,
                boundary=boundary,
                nan_treatment=nan_treatment,
                normalize_kernel=normalize_kernel,
                preserve_nan=preserve_nan,
            )

        if preserve_nan:
            assert np.isnan(z[1])

        x = np.nan_to_num(z)
        z = np.nan_to_num(z)

        if boundary is None:
            assert np.all(z == np.array([0.0, 0.0, 0.0], dtype=">f8"))
        else:
            assert np.all(z == x)

    @pytest.mark.parametrize(
        ("boundary", "nan_treatment", "normalize_kernel", "preserve_nan"),
        itertools.product(
            BOUNDARY_OPTIONS,
            NANHANDLING_OPTIONS,
            NORMALIZE_OPTIONS,
            PRESERVE_NAN_OPTIONS,
        ),
    )
    def test_uniform_3_withnan(
        self, boundary, nan_treatment, normalize_kernel, preserve_nan
    ):
        """
        Test that the different modes are producing the correct results using
        a uniform kernel with three elements. This version includes a NaN
        value in the original array.
        """

        x = np.array([1.0, np.nan, 3.0], dtype=">f8")

        y = np.array([1.0, 1.0, 1.0], dtype=">f8")

        z = convolve(
            x,
            y,
            boundary=boundary,
            nan_treatment=nan_treatment,
            normalize_kernel=normalize_kernel,
            preserve_nan=preserve_nan,
        )

        if preserve_nan:
            assert np.isnan(z[1])

        z = np.nan_to_num(z)

        # boundary, nan_treatment, normalize_kernel
        rslt = {
            (None, "interpolate", True): [0, 2, 0],
            (None, "interpolate", False): [0, 6, 0],
            (None, "fill", True): [0, 4 / 3.0, 0],
            (None, "fill", False): [0, 4, 0],
            ("fill", "interpolate", True): [1 / 2.0, 2, 3 / 2.0],
            ("fill", "interpolate", False): [3 / 2.0, 6, 9 / 2.0],
            ("fill", "fill", True): [1 / 3.0, 4 / 3.0, 3 / 3.0],
            ("fill", "fill", False): [1, 4, 3],
            ("wrap", "interpolate", True): [2, 2, 2],
            ("wrap", "interpolate", False): [6, 6, 6],
            ("wrap", "fill", True): [4 / 3.0, 4 / 3.0, 4 / 3.0],
            ("wrap", "fill", False): [4, 4, 4],
            ("extend", "interpolate", True): [1, 2, 3],
            ("extend", "interpolate", False): [3, 6, 9],
            ("extend", "fill", True): [2 / 3.0, 4 / 3.0, 6 / 3.0],
            ("extend", "fill", False): [2, 4, 6],
        }[boundary, nan_treatment, normalize_kernel]
        if preserve_nan:
            rslt[1] = 0

        assert_array_almost_equal_nulp(z, np.array(rslt, dtype=">f8"), 10)

    @pytest.mark.parametrize(
        ("boundary", "normalize_kernel"),
        itertools.product(BOUNDARY_OPTIONS, NORMALIZE_OPTIONS),
    )
    def test_zero_sum_kernel(self, boundary, normalize_kernel):
        """
        Test that convolve works correctly with zero sum kernels.
        """

        if normalize_kernel:
            pytest.xfail("You can't normalize by a zero sum kernel")

        x = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        y = [-1, -1, -1, -1, 8, -1, -1, -1, -1]
        assert np.isclose(sum(y), 0, atol=1e-8)

        z = convolve(x, y, boundary=boundary, normalize_kernel=normalize_kernel)

        # boundary, normalize_kernel == False
        rslt = {
            (None): [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            "fill": [-6.0, -3.0, -1.0, 0.0, 0.0, 10.0, 21.0, 33.0, 46.0],
            "wrap": [-36.0, -27.0, -18.0, -9.0, 0.0, 9.0, 18.0, 27.0, 36.0],
            "extend": [-10.0, -6.0, -3.0, -1.0, 0.0, 1.0, 3.0, 6.0, 10.0],
        }[boundary]

        assert_array_almost_equal_nulp(z, np.array(rslt, dtype=">f8"), 10)

    @pytest.mark.parametrize(
        ("boundary", "normalize_kernel"),
        itertools.product(BOUNDARY_OPTIONS, NORMALIZE_OPTIONS),
    )
    def test_int_masked_kernel(self, boundary, normalize_kernel):
        """
        Test that convolve works correctly with integer masked kernels.
        """

        if normalize_kernel:
            pytest.xfail("You can't normalize by a zero sum kernel")

        x = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        y = ma.array(
            [-1, -1, -1, -1, 8, -1, -1, -1, -1],
            mask=[1, 0, 0, 0, 0, 0, 0, 0, 0],
            fill_value=0.0,
        )

        z = convolve(x, y, boundary=boundary, normalize_kernel=normalize_kernel)

        # boundary, normalize_kernel == False
        rslt = {
            (None): [0.0, 0.0, 0.0, 0.0, 9.0, 0.0, 0.0, 0.0, 0.0],
            "fill": [-1.0, 3.0, 6.0, 8.0, 9.0, 10.0, 21.0, 33.0, 46.0],
            "wrap": [-31.0, -21.0, -11.0, -1.0, 9.0, 10.0, 20.0, 30.0, 40.0],
            "extend": [-5.0, 0.0, 4.0, 7.0, 9.0, 10.0, 12.0, 15.0, 19.0],
        }[boundary]

        assert_array_almost_equal_nulp(z, np.array(rslt, dtype=">f8"), 10)

    @pytest.mark.parametrize("preserve_nan", PRESERVE_NAN_OPTIONS)
    def test_int_masked_array(self, preserve_nan):
        """
        Test that convolve works correctly with integer masked arrays.
        """

        x = ma.array([3, 5, 7, 11, 13], mask=[0, 0, 1, 0, 0], fill_value=0.0)
        y = np.array([1.0, 1.0, 1.0], dtype=">f8")

        z = convolve(x, y, preserve_nan=preserve_nan)

        if preserve_nan:
            assert np.isnan(z[2])
            z[2] = 8

        assert_array_almost_equal_nulp(z, (8 / 3.0, 4, 8, 12, 8), 10)


class TestConvolve2D:
    def test_list(self):
        """
        Test that convolve works correctly when inputs are lists
        """
        x = [[1, 1, 1], [1, 1, 1], [1, 1, 1]]

        z = convolve(x, x, boundary="fill", fill_value=1, normalize_kernel=True)
        assert_array_almost_equal_nulp(z, x, 10)
        z = convolve(x, x, boundary="fill", fill_value=1, normalize_kernel=False)
        assert_array_almost_equal_nulp(z, np.array(x, float) * 9, 10)

    @pytest.mark.parametrize(("dtype_array", "dtype_kernel"), VALID_DTYPE_MATRIX)
    def test_dtype(self, dtype_array, dtype_kernel):
        """
        Test that 32- and 64-bit floats are correctly handled
        """

        x = np.array(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], dtype=dtype_array
        )

        y = np.array(
            [[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]], dtype=dtype_kernel
        )

        z = convolve(x, y)

        assert x.dtype == z.dtype

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_unity_1x1_none(self, boundary):
        """
        Test that a 1x1 unit kernel returns the same array
        """

        x = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], dtype=">f8")

        y = np.array([[1.0]], dtype=">f8")

        z = convolve(x, y, boundary=boundary)

        assert np.all(z == x)

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_unity_3x3(self, boundary):
        """
        Test that a 3x3 unit kernel returns the same array (except when
        boundary is None).
        """

        x = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], dtype=">f8")

        y = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]], dtype=">f8")

        z = convolve(x, y, boundary=boundary)

        if boundary is None:
            assert np.all(
                z
                == np.array(
                    [[0.0, 0.0, 0.0], [0.0, 5.0, 0.0], [0.0, 0.0, 0.0]], dtype=">f8"
                )
            )
        else:
            assert np.all(z == x)

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_uniform_3x3(self, boundary):
        """
        Test that the different modes are producing the correct results using
        a 3x3 uniform kernel.
        """

        x = np.array([[0.0, 0.0, 3.0], [1.0, 0.0, 0.0], [0.0, 2.0, 0.0]], dtype=">f8")

        y = np.array([[1.0, 1.0, 1.0], [1.0, 1.0, 1.0], [1.0, 1.0, 1.0]], dtype=">f8")

        z = convolve(x, y, boundary=boundary, normalize_kernel=False)

        if boundary is None:
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[0.0, 0.0, 0.0], [0.0, 6.0, 0.0], [0.0, 0.0, 0.0]], dtype=">f8"
                ),
                10,
            )
        elif boundary == "fill":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[1.0, 4.0, 3.0], [3.0, 6.0, 5.0], [3.0, 3.0, 2.0]], dtype=">f8"
                ),
                10,
            )
        elif boundary == "wrap":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[6.0, 6.0, 6.0], [6.0, 6.0, 6.0], [6.0, 6.0, 6.0]], dtype=">f8"
                ),
                10,
            )
        else:
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[2.0, 7.0, 12.0], [4.0, 6.0, 8.0], [6.0, 5.0, 4.0]], dtype=">f8"
                ),
                10,
            )

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_unity_3x3_withnan(self, boundary):
        """
        Test that a 3x3 unit kernel returns the same array (except when
        boundary is None). This version includes a NaN value in the original
        array.
        """

        x = np.array(
            [[1.0, 2.0, 3.0], [4.0, np.nan, 6.0], [7.0, 8.0, 9.0]], dtype=">f8"
        )

        y = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]], dtype=">f8")

        z = convolve(x, y, boundary=boundary, nan_treatment="fill", preserve_nan=True)

        assert np.isnan(z[1, 1])
        x = np.nan_to_num(z)
        z = np.nan_to_num(z)

        if boundary is None:
            assert np.all(
                z
                == np.array(
                    [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], dtype=">f8"
                )
            )
        else:
            assert np.all(z == x)

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_uniform_3x3_withnanfilled(self, boundary):
        """
        Test that the different modes are producing the correct results using
        a 3x3 uniform kernel. This version includes a NaN value in the
        original array.
        """

        x = np.array(
            [[0.0, 0.0, 4.0], [1.0, np.nan, 0.0], [0.0, 3.0, 0.0]], dtype=">f8"
        )

        y = np.array([[1.0, 1.0, 1.0], [1.0, 1.0, 1.0], [1.0, 1.0, 1.0]], dtype=">f8")

        z = convolve(
            x, y, boundary=boundary, nan_treatment="fill", normalize_kernel=False
        )

        if boundary is None:
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[0.0, 0.0, 0.0], [0.0, 8.0, 0.0], [0.0, 0.0, 0.0]], dtype=">f8"
                ),
                10,
            )
        elif boundary == "fill":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[1.0, 5.0, 4.0], [4.0, 8.0, 7.0], [4.0, 4.0, 3.0]], dtype=">f8"
                ),
                10,
            )
        elif boundary == "wrap":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[8.0, 8.0, 8.0], [8.0, 8.0, 8.0], [8.0, 8.0, 8.0]], dtype=">f8"
                ),
                10,
            )
        elif boundary == "extend":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[2.0, 9.0, 16.0], [5.0, 8.0, 11.0], [8.0, 7.0, 6.0]], dtype=">f8"
                ),
                10,
            )
        else:
            raise ValueError("Invalid boundary specification")

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_uniform_3x3_withnaninterped(self, boundary):
        """
        Test that the different modes are producing the correct results using
        a 3x3 uniform kernel. This version includes a NaN value in the
        original array.
        """

        x = np.array(
            [[0.0, 0.0, 4.0], [1.0, np.nan, 0.0], [0.0, 3.0, 0.0]], dtype=">f8"
        )

        y = np.array([[1.0, 1.0, 1.0], [1.0, 1.0, 1.0], [1.0, 1.0, 1.0]], dtype=">f8")

        z = convolve(
            x, y, boundary=boundary, nan_treatment="interpolate", normalize_kernel=True
        )

        if boundary is None:
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]], dtype=">f8"
                ),
                10,
            )
        elif boundary == "fill":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [1.0 / 8, 5.0 / 8, 4.0 / 8],
                        [4.0 / 8, 8.0 / 8, 7.0 / 8],
                        [4.0 / 8, 4.0 / 8, 3.0 / 8],
                    ],
                    dtype=">f8",
                ),
                10,
            )
        elif boundary == "wrap":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[1.0, 1.0, 1.0], [1.0, 1.0, 1.0], [1.0, 1.0, 1.0]], dtype=">f8"
                ),
                10,
            )
        elif boundary == "extend":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [2.0 / 8, 9.0 / 8, 16.0 / 8],
                        [5.0 / 8, 8.0 / 8, 11.0 / 8],
                        [8.0 / 8, 7.0 / 8, 6.0 / 8],
                    ],
                    dtype=">f8",
                ),
                10,
            )
        else:
            raise ValueError("Invalid boundary specification")

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_non_normalized_kernel_2D(self, boundary):
        x = np.array([[0.0, 0.0, 4.0], [1.0, 2.0, 0.0], [0.0, 3.0, 0.0]], dtype="float")

        y = np.array(
            [[1.0, -1.0, 1.0], [-1.0, 0.0, -1.0], [1.0, -1.0, 1.0]], dtype="float"
        )

        z = convolve(
            x, y, boundary=boundary, nan_treatment="fill", normalize_kernel=False
        )

        if boundary is None:
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], dtype="float"
                ),
                10,
            )
        elif boundary == "fill":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[1.0, -5.0, 2.0], [1.0, 0.0, -3.0], [-2.0, -1.0, -1.0]],
                    dtype="float",
                ),
                10,
            )
        elif boundary == "wrap":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[0.0, -8.0, 6.0], [5.0, 0.0, -4.0], [2.0, 3.0, -4.0]],
                    dtype="float",
                ),
                10,
            )
        elif boundary == "extend":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [[2.0, -1.0, -2.0], [0.0, 0.0, 1.0], [2.0, -4.0, 2.0]],
                    dtype="float",
                ),
                10,
            )
        else:
            raise ValueError("Invalid boundary specification")


class TestConvolve3D:
    def test_list(self):
        """
        Test that convolve works correctly when inputs are lists
        """
        x = [
            [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
            [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
            [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
        ]

        z = convolve(x, x, boundary="fill", fill_value=1, normalize_kernel=False)
        assert_array_almost_equal_nulp(z / 27, x, 10)

    @pytest.mark.parametrize(("dtype_array", "dtype_kernel"), VALID_DTYPE_MATRIX)
    def test_dtype(self, dtype_array, dtype_kernel):
        """
        Test that 32- and 64-bit floats are correctly handled
        """

        x = np.array(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], dtype=dtype_array
        )

        y = np.array(
            [[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]], dtype=dtype_kernel
        )

        z = convolve(x, y)

        assert x.dtype == z.dtype

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_unity_1x1x1_none(self, boundary):
        """
        Test that a 1x1x1 unit kernel returns the same array
        """

        x = np.array(
            [
                [[1.0, 2.0, 1.0], [2.0, 3.0, 1.0], [3.0, 2.0, 5.0]],
                [[4.0, 3.0, 1.0], [5.0, 0.0, 2.0], [6.0, 1.0, 1.0]],
                [[7.0, 0.0, 2.0], [8.0, 2.0, 3.0], [9.0, 2.0, 2.0]],
            ],
            dtype=">f8",
        )

        y = np.array([[[1.0]]], dtype=">f8")

        z = convolve(x, y, boundary=boundary)

        assert np.all(z == x)

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_unity_3x3x3(self, boundary):
        """
        Test that a 3x3x3 unit kernel returns the same array (except when
        boundary is None).
        """

        x = np.array(
            [
                [[1.0, 2.0, 1.0], [2.0, 3.0, 1.0], [3.0, 2.0, 5.0]],
                [[4.0, 3.0, 1.0], [5.0, 3.0, 2.0], [6.0, 1.0, 1.0]],
                [[7.0, 0.0, 2.0], [8.0, 2.0, 3.0], [9.0, 2.0, 2.0]],
            ],
            dtype=">f8",
        )

        y = np.zeros((3, 3, 3), dtype=">f8")
        y[1, 1, 1] = 1.0

        z = convolve(x, y, boundary=boundary)

        if boundary is None:
            assert np.all(
                z
                == np.array(
                    [
                        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                        [[0.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 0.0]],
                        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                    ],
                    dtype=">f8",
                )
            )
        else:
            assert np.all(z == x)

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_uniform_3x3x3(self, boundary):
        """
        Test that the different modes are producing the correct results using
        a 3x3 uniform kernel.
        """

        x = np.array(
            [
                [[1.0, 2.0, 1.0], [2.0, 3.0, 1.0], [3.0, 2.0, 5.0]],
                [[4.0, 3.0, 1.0], [5.0, 3.0, 2.0], [6.0, 1.0, 1.0]],
                [[7.0, 0.0, 2.0], [8.0, 2.0, 3.0], [9.0, 2.0, 2.0]],
            ],
            dtype=">f8",
        )

        y = np.ones((3, 3, 3), dtype=">f8")

        z = convolve(x, y, boundary=boundary, normalize_kernel=False)

        if boundary is None:
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                        [[0.0, 0.0, 0.0], [0.0, 81.0, 0.0], [0.0, 0.0, 0.0]],
                        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                    ],
                    dtype=">f8",
                ),
                10,
            )
        elif boundary == "fill":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [[23.0, 28.0, 16.0], [35.0, 46.0, 25.0], [25.0, 34.0, 18.0]],
                        [[40.0, 50.0, 23.0], [63.0, 81.0, 36.0], [46.0, 60.0, 27.0]],
                        [[32.0, 40.0, 16.0], [50.0, 61.0, 22.0], [36.0, 44.0, 16.0]],
                    ],
                    dtype=">f8",
                ),
                10,
            )
        elif boundary == "wrap":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [[81.0, 81.0, 81.0], [81.0, 81.0, 81.0], [81.0, 81.0, 81.0]],
                        [[81.0, 81.0, 81.0], [81.0, 81.0, 81.0], [81.0, 81.0, 81.0]],
                        [[81.0, 81.0, 81.0], [81.0, 81.0, 81.0], [81.0, 81.0, 81.0]],
                    ],
                    dtype=">f8",
                ),
                10,
            )
        else:
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [[65.0, 54.0, 43.0], [75.0, 66.0, 57.0], [85.0, 78.0, 71.0]],
                        [[96.0, 71.0, 46.0], [108.0, 81.0, 54.0], [120.0, 91.0, 62.0]],
                        [
                            [127.0, 88.0, 49.0],
                            [141.0, 96.0, 51.0],
                            [155.0, 104.0, 53.0],
                        ],
                    ],
                    dtype=">f8",
                ),
                10,
            )

    @pytest.mark.parametrize(
        ("boundary", "nan_treatment"),
        itertools.product(BOUNDARY_OPTIONS, NANHANDLING_OPTIONS),
    )
    def test_unity_3x3x3_withnan(self, boundary, nan_treatment):
        """
        Test that a 3x3x3 unit kernel returns the same array (except when
        boundary is None). This version includes a NaN value in the original
        array.
        """

        x = np.array(
            [
                [[1.0, 2.0, 1.0], [2.0, 3.0, 1.0], [3.0, 2.0, 5.0]],
                [[4.0, 3.0, 1.0], [5.0, np.nan, 2.0], [6.0, 1.0, 1.0]],
                [[7.0, 0.0, 2.0], [8.0, 2.0, 3.0], [9.0, 2.0, 2.0]],
            ],
            dtype=">f8",
        )

        y = np.zeros((3, 3, 3), dtype=">f8")
        y[1, 1, 1] = 1.0

        z = convolve(
            x, y, boundary=boundary, nan_treatment=nan_treatment, preserve_nan=True
        )

        assert np.isnan(z[1, 1, 1])
        x = np.nan_to_num(z)
        z = np.nan_to_num(z)

        if boundary is None:
            assert np.all(
                z
                == np.array(
                    [
                        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                    ],
                    dtype=">f8",
                )
            )
        else:
            assert np.all(z == x)

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_uniform_3x3x3_withnan_filled(self, boundary):
        """
        Test that the different modes are producing the correct results using
        a 3x3 uniform kernel. This version includes a NaN value in the
        original array.
        """

        x = np.array(
            [
                [[1.0, 2.0, 1.0], [2.0, 3.0, 1.0], [3.0, 2.0, 5.0]],
                [[4.0, 3.0, 1.0], [5.0, np.nan, 2.0], [6.0, 1.0, 1.0]],
                [[7.0, 0.0, 2.0], [8.0, 2.0, 3.0], [9.0, 2.0, 2.0]],
            ],
            dtype=">f8",
        )

        y = np.ones((3, 3, 3), dtype=">f8")

        z = convolve(
            x, y, boundary=boundary, nan_treatment="fill", normalize_kernel=False
        )

        if boundary is None:
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                        [[0.0, 0.0, 0.0], [0.0, 78.0, 0.0], [0.0, 0.0, 0.0]],
                        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                    ],
                    dtype=">f8",
                ),
                10,
            )
        elif boundary == "fill":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [[20.0, 25.0, 13.0], [32.0, 43.0, 22.0], [22.0, 31.0, 15.0]],
                        [[37.0, 47.0, 20.0], [60.0, 78.0, 33.0], [43.0, 57.0, 24.0]],
                        [[29.0, 37.0, 13.0], [47.0, 58.0, 19.0], [33.0, 41.0, 13.0]],
                    ],
                    dtype=">f8",
                ),
                10,
            )
        elif boundary == "wrap":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [[78.0, 78.0, 78.0], [78.0, 78.0, 78.0], [78.0, 78.0, 78.0]],
                        [[78.0, 78.0, 78.0], [78.0, 78.0, 78.0], [78.0, 78.0, 78.0]],
                        [[78.0, 78.0, 78.0], [78.0, 78.0, 78.0], [78.0, 78.0, 78.0]],
                    ],
                    dtype=">f8",
                ),
                10,
            )
        elif boundary == "extend":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [[62.0, 51.0, 40.0], [72.0, 63.0, 54.0], [82.0, 75.0, 68.0]],
                        [[93.0, 68.0, 43.0], [105.0, 78.0, 51.0], [117.0, 88.0, 59.0]],
                        [
                            [124.0, 85.0, 46.0],
                            [138.0, 93.0, 48.0],
                            [152.0, 101.0, 50.0],
                        ],
                    ],
                    dtype=">f8",
                ),
                10,
            )
        else:
            raise ValueError("Invalid Boundary Option")

    @pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
    def test_uniform_3x3x3_withnan_interped(self, boundary):
        """
        Test that the different modes are producing the correct results using
        a 3x3 uniform kernel. This version includes a NaN value in the
        original array.
        """

        x = np.array(
            [
                [[1.0, 2.0, 1.0], [2.0, 3.0, 1.0], [3.0, 2.0, 5.0]],
                [[4.0, 3.0, 1.0], [5.0, np.nan, 2.0], [6.0, 1.0, 1.0]],
                [[7.0, 0.0, 2.0], [8.0, 2.0, 3.0], [9.0, 2.0, 2.0]],
            ],
            dtype=">f8",
        )

        y = np.ones((3, 3, 3), dtype=">f8")

        z = convolve(
            x, y, boundary=boundary, nan_treatment="interpolate", normalize_kernel=True
        )

        kernsum = y.sum() - 1  # one nan is missing
        mid = x[np.isfinite(x)].sum() / kernsum

        if boundary is None:
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                        [[0.0, 0.0, 0.0], [0.0, 78.0, 0.0], [0.0, 0.0, 0.0]],
                        [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                    ],
                    dtype=">f8",
                )
                / kernsum,
                10,
            )
        elif boundary == "fill":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [[20.0, 25.0, 13.0], [32.0, 43.0, 22.0], [22.0, 31.0, 15.0]],
                        [[37.0, 47.0, 20.0], [60.0, 78.0, 33.0], [43.0, 57.0, 24.0]],
                        [[29.0, 37.0, 13.0], [47.0, 58.0, 19.0], [33.0, 41.0, 13.0]],
                    ],
                    dtype=">f8",
                )
                / kernsum,
                10,
            )
        elif boundary == "wrap":
            assert_array_almost_equal_nulp(z, np.tile(mid.astype(">f8"), [3, 3, 3]), 10)
        elif boundary == "extend":
            assert_array_almost_equal_nulp(
                z,
                np.array(
                    [
                        [[62.0, 51.0, 40.0], [72.0, 63.0, 54.0], [82.0, 75.0, 68.0]],
                        [[93.0, 68.0, 43.0], [105.0, 78.0, 51.0], [117.0, 88.0, 59.0]],
                        [
                            [124.0, 85.0, 46.0],
                            [138.0, 93.0, 48.0],
                            [152.0, 101.0, 50.0],
                        ],
                    ],
                    dtype=">f8",
                )
                / kernsum,
                10,
            )
        else:
            raise ValueError("Invalid Boundary Option")


@pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
def test_asymmetric_kernel(boundary):
    """
    Regression test for #6264: make sure that asymmetric convolution
    functions go the right direction
    """

    x = np.array([3.0, 0.0, 1.0], dtype=">f8")

    y = np.array([1, 2, 3], dtype=">f8")

    z = convolve(x, y, boundary=boundary, normalize_kernel=False)

    if boundary == "fill":
        assert_array_almost_equal_nulp(z, np.array([6.0, 10.0, 2.0], dtype="float"), 10)
    elif boundary is None:
        assert_array_almost_equal_nulp(z, np.array([0.0, 10.0, 0.0], dtype="float"), 10)
    elif boundary == "extend":
        assert_array_almost_equal_nulp(
            z, np.array([15.0, 10.0, 3.0], dtype="float"), 10
        )
    elif boundary == "wrap":
        assert_array_almost_equal_nulp(z, np.array([9.0, 10.0, 5.0], dtype="float"), 10)


@pytest.mark.parametrize("ndims", (1, 2, 3))
def test_convolution_consistency(ndims):
    np.random.seed(0)
    array = np.random.randn(*([3] * ndims))
    np.random.seed(0)
    kernel = np.random.rand(*([3] * ndims))

    conv_f = convolve_fft(array, kernel, boundary="fill")
    conv_d = convolve(array, kernel, boundary="fill")

    assert_array_almost_equal_nulp(conv_f, conv_d, 30)


def test_astropy_convolution_against_numpy():
    x = np.array([1, 2, 3])
    y = np.array([5, 4, 3, 2, 1])

    assert_array_almost_equal(
        np.convolve(y, x, "same"), convolve(y, x, normalize_kernel=False)
    )
    assert_array_almost_equal(
        np.convolve(y, x, "same"), convolve_fft(y, x, normalize_kernel=False)
    )


@pytest.mark.skipif(not HAS_SCIPY, reason="Requires scipy")
def test_astropy_convolution_against_scipy():
    from scipy.signal import fftconvolve

    x = np.array([1, 2, 3])
    y = np.array([5, 4, 3, 2, 1])

    assert_array_almost_equal(
        fftconvolve(y, x, "same"), convolve(y, x, normalize_kernel=False)
    )
    assert_array_almost_equal(
        fftconvolve(y, x, "same"), convolve_fft(y, x, normalize_kernel=False)
    )


@pytest.mark.skipif(not HAS_PANDAS, reason="Requires pandas")
def test_regression_6099():
    import pandas

    wave = np.array(np.linspace(5000, 5100, 10))
    boxcar = 3
    nonseries_result = convolve(wave, np.ones((boxcar,)) / boxcar)

    wave_series = pandas.Series(wave)
    series_result = convolve(wave_series, np.ones((boxcar,)) / boxcar)

    assert_array_almost_equal(nonseries_result, series_result)


def test_invalid_array_convolve():
    kernel = np.ones(3) / 3.0

    with pytest.raises(TypeError):
        convolve("glork", kernel)


@pytest.mark.parametrize("boundary", BOUNDARY_OPTIONS)
def test_non_square_kernel_asymmetric(boundary):
    # Regression test for a bug that occurred when using non-square kernels in
    # 2D when using boundary=None
    kernel = np.array([[1, 2, 3, 2, 1], [0, 1, 2, 1, 0], [0, 0, 0, 0, 0]])
    image = np.zeros((13, 13))
    image[6, 6] = 1
    result = convolve(image, kernel, normalize_kernel=False, boundary=boundary)
    assert_allclose(result[5:8, 4:9], kernel)


@pytest.mark.parametrize(
    ("boundary", "normalize_kernel"),
    itertools.product(BOUNDARY_OPTIONS, NORMALIZE_OPTIONS),
)
def test_uninterpolated_nan_regions(boundary, normalize_kernel):
    # Issue #8086
    # Test NaN interpolation of contiguous NaN regions with kernels of size
    # identical and greater than that of the region of NaN values.

    # Test case: kernel.shape == NaN_region.shape
    kernel = Gaussian2DKernel(1, 5, 5)
    nan_centroid = np.full(kernel.shape, np.nan)
    image = np.pad(
        nan_centroid, pad_width=kernel.shape[0] * 2, mode="constant", constant_values=1
    )
    with pytest.warns(
        AstropyUserWarning,
        match=r"nan_treatment='interpolate', however, NaN values detected "
        r"post convolution. A contiguous region of NaN values, larger "
        r"than the kernel size, are present in the input array. "
        r"Increase the kernel size to avoid this.",
    ):
        result = convolve(
            image,
            kernel,
            boundary=boundary,
            nan_treatment="interpolate",
            normalize_kernel=normalize_kernel,
        )
        assert np.any(np.isnan(result))

    # Test case: kernel.shape > NaN_region.shape
    nan_centroid = np.full(
        (kernel.shape[0] - 1, kernel.shape[1] - 1), np.nan
    )  # 1 smaller than kerenel
    image = np.pad(
        nan_centroid, pad_width=kernel.shape[0] * 2, mode="constant", constant_values=1
    )
    result = convolve(
        image,
        kernel,
        boundary=boundary,
        nan_treatment="interpolate",
        normalize_kernel=normalize_kernel,
    )
    assert ~np.any(np.isnan(result))  # Note: negation


def test_regressiontest_issue9168():
    """
    Issue #9168 pointed out that kernels can be (unitless) quantities, which
    leads to crashes when inplace modifications are made to arrays in
    convolve/convolve_fft, so we now strip the quantity aspects off of kernels.
    """

    x = np.array(
        [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
    )

    kernel_fwhm = 1 * u.arcsec
    pixel_size = 1 * u.arcsec

    kernel = Gaussian2DKernel(x_stddev=kernel_fwhm / pixel_size)

    convolve_fft(x, kernel, boundary="fill", fill_value=np.nan, preserve_nan=True)
    convolve(x, kernel, boundary="fill", fill_value=np.nan, preserve_nan=True)


def test_convolve_nan_zero_sum_kernel():
    with pytest.raises(
        ValueError,
        match=(
            "Setting nan_treatment='interpolate' "
            "requires the kernel to be normalized, but the "
            "input kernel has a sum close to zero. For a "
            "zero-sum kernel and data with NaNs, set "
            "nan_treatment='fill'."
        ),
    ):
        convolve([1, np.nan, 3], [-1, 2, -1], normalize_kernel=False)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from asdf.versioning import AsdfSpec
from numpy.testing import assert_array_equal

from astropy import time
from astropy import units as u
from astropy.coordinates import EarthLocation
from astropy.io.misc.asdf.types import AstropyAsdfType
from astropy.units import Quantity

__all__ = ["TimeType"]

_guessable_formats = {"iso", "byear", "jyear", "yday"}

_astropy_format_to_asdf_format = {
    "isot": "iso",
    "byear_str": "byear",
    "jyear_str": "jyear",
}


def _assert_earthlocation_equal(a, b):
    assert_array_equal(a.x, b.x)
    assert_array_equal(a.y, b.y)
    assert_array_equal(a.z, b.z)
    assert_array_equal(a.lat, b.lat)
    assert_array_equal(a.lon, b.lon)


class TimeType(AstropyAsdfType):
    name = "time/time"
    version = "1.1.0"
    supported_versions = ["1.0.0", AsdfSpec(">=1.1.0")]
    types = ["astropy.time.core.Time"]
    requires = ["astropy"]

    @classmethod
    def to_tree(cls, node, ctx):
        fmt = node.format

        if fmt == "byear":
            node = time.Time(node, format="byear_str")

        elif fmt == "jyear":
            node = time.Time(node, format="jyear_str")

        elif fmt in ("fits", "datetime", "plot_date"):
            node = time.Time(node, format="isot")

        fmt = node.format

        fmt = _astropy_format_to_asdf_format.get(fmt, fmt)

        guessable_format = fmt in _guessable_formats

        if node.scale == "utc" and guessable_format and node.isscalar:
            return node.value

        d = {"value": node.value}

        if not guessable_format:
            d["format"] = fmt

        if node.scale != "utc":
            d["scale"] = node.scale

        if node.location is not None:
            x, y, z = node.location.x, node.location.y, node.location.z
            # Preserve backwards compatibility for writing the old schema
            # This allows WCS to test backwards compatibility with old frames
            # This code does get tested in CI, but we don't run a coverage test
            if cls.version == "1.0.0":  # pragma: no cover
                unit = node.location.unit
                d["location"] = {"x": x.value, "y": y.value, "z": z.value, "unit": unit}
            else:
                d["location"] = {
                    # It seems like EarthLocations can be represented either in
                    # terms of Cartesian coordinates or latitude and longitude, so
                    # we rather arbitrarily choose the former for our representation
                    "x": x,
                    "y": y,
                    "z": z,
                }

        return d

    @classmethod
    def from_tree(cls, node, ctx):
        if isinstance(node, (str, list, np.ndarray)):
            t = time.Time(node)
            fmt = _astropy_format_to_asdf_format.get(t.format, t.format)
            if fmt not in _guessable_formats:
                raise ValueError(f"Invalid time '{node}'")
            return t

        value = node["value"]
        fmt = node.get("format")
        scale = node.get("scale")
        location = node.get("location")
        if location is not None:
            unit = location.get("unit", u.m)
            # This ensures that we can read the v.1.0.0 schema and convert it
            # to the new EarthLocation object, which expects Quantity components
            for comp in ["x", "y", "z"]:
                if not isinstance(location[comp], Quantity):
                    location[comp] = Quantity(location[comp], unit=unit)
            location = EarthLocation.from_geocentric(
                location["x"], location["y"], location["z"]
            )

        return time.Time(value, format=fmt, scale=scale, location=location)

    @classmethod
    def assert_equal(cls, old, new):
        assert old.format == new.format
        assert old.scale == new.scale
        if isinstance(old.location, EarthLocation):
            assert isinstance(new.location, EarthLocation)
            _assert_earthlocation_equal(old.location, new.location)
        else:
            assert old.location == new.location

        assert_array_equal(old, new)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal
from packaging.version import Version

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.nddata import (
    CCDData,
    Cutout2D,
    NoOverlapError,
    PartialOverlapError,
    add_array,
    extract_array,
    overlap_slices,
    subpixel_indices,
)
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs import WCS, Sip
from astropy.wcs.utils import proj_plane_pixel_area

test_positions = [
    (10.52, 3.12),
    (5.62, 12.97),
    (31.33, 31.77),
    (0.46, 0.94),
    (20.45, 12.12),
    (42.24, 24.42),
]

test_position_indices = [(0, 3), (0, 2), (4, 1), (4, 2), (4, 3), (3, 4)]

test_slices = [
    slice(10.52, 3.12),
    slice(5.62, 12.97),
    slice(31.33, 31.77),
    slice(0.46, 0.94),
    slice(20.45, 12.12),
    slice(42.24, 24.42),
]

subsampling = 5

test_pos_bad = [(-1, -4), (-2, 0), (6, 2), (6, 6)]
test_nonfinite_positions = [
    (np.nan, np.nan),
    (np.inf, np.inf),
    (1, np.nan),
    (np.nan, 2),
    (2, -np.inf),
    (-np.inf, 3),
]


def test_slices_different_dim():
    """Overlap from arrays with different number of dim is undefined."""
    with pytest.raises(ValueError, match=".*the same number of dimensions.*"):
        overlap_slices((4, 5, 6), (1, 2), (0, 0))


def test_slices_pos_different_dim():
    """Position must have same dim as arrays."""
    with pytest.raises(ValueError, match=".*the same number of dimensions.*"):
        overlap_slices((4, 5), (1, 2), (0, 0, 3))


@pytest.mark.parametrize("pos", test_pos_bad)
def test_slices_no_overlap(pos):
    """If there is no overlap between arrays, an error should be raised."""
    with pytest.raises(NoOverlapError):
        overlap_slices((5, 5), (2, 2), pos)


def test_slices_partial_overlap():
    """Compute a slice for partially overlapping arrays."""
    temp = overlap_slices((5,), (3,), (0,))
    assert temp == ((slice(0, 2, None),), (slice(1, 3, None),))

    temp = overlap_slices((5,), (3,), (0,), mode="partial")
    assert temp == ((slice(0, 2, None),), (slice(1, 3, None),))

    for pos in [0, 4]:
        with pytest.raises(
            PartialOverlapError, match=".*Arrays overlap only partially.*"
        ):
            temp = overlap_slices((5,), (3,), (pos,), mode="strict")


def test_slices_edges():
    """
    Test overlap_slices when extracting along edges.
    """

    slc_lg, slc_sm = overlap_slices((10, 10), (3, 3), (1, 1), mode="strict")
    assert slc_lg[0].start == slc_lg[1].start == 0
    assert slc_lg[0].stop == slc_lg[1].stop == 3
    assert slc_sm[0].start == slc_sm[1].start == 0
    assert slc_sm[0].stop == slc_sm[1].stop == 3

    slc_lg, slc_sm = overlap_slices((10, 10), (3, 3), (8, 8), mode="strict")
    assert slc_lg[0].start == slc_lg[1].start == 7
    assert slc_lg[0].stop == slc_lg[1].stop == 10
    assert slc_sm[0].start == slc_sm[1].start == 0
    assert slc_sm[0].stop == slc_sm[1].stop == 3

    # test (0, 0) shape
    slc_lg, slc_sm = overlap_slices((10, 10), (0, 0), (0, 0))
    assert slc_lg[0].start == slc_lg[0].stop == 0
    assert slc_lg[1].start == slc_lg[1].stop == 0
    assert slc_sm[0].start == slc_sm[0].stop == 0
    assert slc_sm[1].start == slc_sm[1].stop == 0

    slc_lg, slc_sm = overlap_slices((10, 10), (0, 0), (5, 5))
    assert slc_lg[0].start == slc_lg[0].stop == 5
    assert slc_lg[1].start == slc_lg[1].stop == 5
    assert slc_sm[0].start == slc_sm[0].stop == 0
    assert slc_sm[1].start == slc_sm[1].stop == 0


def test_slices_overlap_wrong_mode():
    """Call overlap_slices with non-existing mode."""
    with pytest.raises(ValueError, match="^Mode can be only.*"):
        overlap_slices((5,), (3,), (0,), mode="full")


@pytest.mark.parametrize("position", test_nonfinite_positions)
def test_slices_nonfinite_position(position):
    """
    A ValueError should be raised if position contains a non-finite
    value.
    """

    with pytest.raises(ValueError):
        overlap_slices((7, 7), (3, 3), position)


def test_extract_array_even_shape_rounding():
    """
    Test overlap_slices (via extract_array) for rounding with an
    even-shaped extraction.
    """

    data = np.arange(10)
    shape = (2,)
    positions_expected = [
        (1.49, (1, 2)),
        (1.5, (1, 2)),
        (1.501, (1, 2)),
        (1.99, (1, 2)),
        (2.0, (1, 2)),
        (2.01, (2, 3)),
        (2.49, (2, 3)),
        (2.5, (2, 3)),
        (2.501, (2, 3)),
        (2.99, (2, 3)),
        (3.0, (2, 3)),
        (3.01, (3, 4)),
    ]

    for pos, exp in positions_expected:
        out = extract_array(data, shape, (pos,), mode="partial")
        assert_array_equal(out, exp)

    # test negative positions
    positions = (-0.99, -0.51, -0.5, -0.49, -0.01, 0)
    exp1 = (-99, 0)
    exp2 = (0, 1)
    expected = [exp1] * 6 + [exp2]

    for pos, exp in zip(positions, expected):
        out = extract_array(data, shape, (pos,), mode="partial", fill_value=-99)
        assert_array_equal(out, exp)


def test_extract_array_odd_shape_rounding():
    """
    Test overlap_slices (via extract_array) for rounding with an
    even-shaped extraction.
    """

    data = np.arange(10)
    shape = (3,)
    positions_expected = [
        (1.49, (0, 1, 2)),
        (1.5, (0, 1, 2)),
        (1.501, (1, 2, 3)),
        (1.99, (1, 2, 3)),
        (2.0, (1, 2, 3)),
        (2.01, (1, 2, 3)),
        (2.49, (1, 2, 3)),
        (2.5, (1, 2, 3)),
        (2.501, (2, 3, 4)),
        (2.99, (2, 3, 4)),
        (3.0, (2, 3, 4)),
        (3.01, (2, 3, 4)),
    ]

    for pos, exp in positions_expected:
        out = extract_array(data, shape, (pos,), mode="partial")
        assert_array_equal(out, exp)

    # test negative positions
    positions = (-0.99, -0.51, -0.5, -0.49, -0.01, 0)
    exp1 = (-99, -99, 0)
    exp2 = (-99, 0, 1)
    expected = [exp1] * 3 + [exp2] * 4

    for pos, exp in zip(positions, expected):
        out = extract_array(data, shape, (pos,), mode="partial", fill_value=-99)
        assert_array_equal(out, exp)


def test_extract_array_wrong_mode():
    """Call extract_array with non-existing mode."""
    with pytest.raises(
        ValueError, match="Valid modes are 'partial', 'trim', and 'strict'."
    ):
        extract_array(np.arange(4), (2,), (0,), mode="full")


def test_extract_array_1d_even():
    """Extract 1 d arrays.

    All dimensions are treated the same, so we can test in 1 dim.
    """
    assert np.all(
        extract_array(np.arange(4), (2,), (0,), fill_value=-99) == np.array([-99, 0])
    )
    for i in [1, 2, 3]:
        assert np.all(extract_array(np.arange(4), (2,), (i,)) == np.array([i - 1, i]))
    assert np.all(
        extract_array(np.arange(4.0), (2,), (4,), fill_value=np.inf)
        == np.array([3, np.inf])
    )


def test_extract_array_1d_odd():
    """Extract 1 d arrays.

    All dimensions are treated the same, so we can test in 1 dim.
    The first few lines test the most error-prone part: Extraction of an
    array on the boundaries.
    Additional tests (e.g. dtype of return array) are done for the last
    case only.
    """
    assert np.all(
        extract_array(np.arange(4), (3,), (-1,), fill_value=-99)
        == np.array([-99, -99, 0])
    )
    assert np.all(
        extract_array(np.arange(4), (3,), (0,), fill_value=-99) == np.array([-99, 0, 1])
    )
    for i in [1, 2]:
        assert np.all(
            extract_array(np.arange(4), (3,), (i,)) == np.array([i - 1, i, i + 1])
        )
    assert np.all(
        extract_array(np.arange(4), (3,), (3,), fill_value=-99) == np.array([2, 3, -99])
    )
    arrayin = np.arange(4.0)
    extracted = extract_array(arrayin, (3,), (4,))
    assert extracted[0] == 3
    assert np.isnan(extracted[1])  # since I cannot use `==` to test for nan
    assert extracted.dtype == arrayin.dtype


def test_extract_array_1d():
    """In 1d, shape can be int instead of tuple"""
    assert np.all(
        extract_array(np.arange(4), 3, (-1,), fill_value=-99) == np.array([-99, -99, 0])
    )
    assert np.all(
        extract_array(np.arange(4), 3, -1, fill_value=-99) == np.array([-99, -99, 0])
    )


def test_extract_Array_float():
    """integer is at bin center"""
    for a in np.arange(2.51, 3.49, 0.1):
        assert np.all(extract_array(np.arange(5), 3, a) == np.array([2, 3, 4]))


def test_extract_array_1d_trim():
    """Extract 1 d arrays.

    All dimensions are treated the same, so we can test in 1 dim.
    """
    assert np.all(extract_array(np.arange(4), (2,), (0,), mode="trim") == np.array([0]))
    for i in [1, 2, 3]:
        assert np.all(
            extract_array(np.arange(4), (2,), (i,), mode="trim") == np.array([i - 1, i])
        )
    assert np.all(
        extract_array(np.arange(4.0), (2,), (4,), mode="trim") == np.array([3])
    )


@pytest.mark.parametrize("mode", ["partial", "trim", "strict"])
def test_extract_array_easy(mode):
    """
    Test extract_array utility function.

    Test by extracting an array of ones out of an array of zeros.
    """
    large_test_array = np.zeros((11, 11))
    small_test_array = np.ones((5, 5))
    large_test_array[3:8, 3:8] = small_test_array
    extracted_array = extract_array(large_test_array, (5, 5), (5, 5), mode=mode)
    assert np.all(extracted_array == small_test_array)


def test_extract_array_return_pos():
    """Check that the return position is calculated correctly.

    The result will differ by mode. All test here are done in 1d because it's
    easier to construct correct test cases.
    """
    large_test_array = np.arange(5, dtype=float)
    for i in np.arange(-1, 6):
        extracted, new_pos = extract_array(
            large_test_array, 3, i, mode="partial", return_position=True
        )
        assert new_pos == (1,)
    # Now check an array with an even number
    for i, expected in zip([1.49, 1.51, 3], [0.49, 0.51, 1]):
        extracted, new_pos = extract_array(
            large_test_array, (2,), (i,), mode="strict", return_position=True
        )
        assert new_pos == (expected,)
    # For mode='trim' the answer actually depends
    for i, expected in zip(np.arange(-1, 6), (-1, 0, 1, 1, 1, 1, 1)):
        extracted, new_pos = extract_array(
            large_test_array, (3,), (i,), mode="trim", return_position=True
        )
        assert new_pos == (expected,)


def test_extract_array_nan_fillvalue():
    if Version(np.__version__) >= Version("1.20"):
        msg = "fill_value cannot be set to np.nan if the input array has"
        with pytest.raises(ValueError, match=msg):
            extract_array(
                np.ones((10, 10), dtype=int), (5, 5), (1, 1), fill_value=np.nan
            )


def test_add_array_odd_shape():
    """
    Test add_array utility function.

    Test by adding an array of ones out of an array of zeros.
    """
    large_test_array = np.zeros((11, 11))
    small_test_array = np.ones((5, 5))
    large_test_array_ref = large_test_array.copy()
    large_test_array_ref[3:8, 3:8] += small_test_array

    added_array = add_array(large_test_array, small_test_array, (5, 5))
    assert np.all(added_array == large_test_array_ref)


def test_add_array_even_shape():
    """
    Test add_array_2D utility function.

    Test by adding an array of ones out of an array of zeros.
    """
    large_test_array = np.zeros((11, 11))
    small_test_array = np.ones((4, 4))
    large_test_array_ref = large_test_array.copy()
    large_test_array_ref[0:2, 0:2] += small_test_array[2:4, 2:4]

    added_array = add_array(large_test_array, small_test_array, (0, 0))
    assert np.all(added_array == large_test_array_ref)


def test_add_array_equal_shape():
    """
    Test add_array_2D utility function.

    Test by adding an array of ones out of an array of zeros.
    """
    large_test_array = np.zeros((11, 11))
    small_test_array = np.ones((11, 11))
    large_test_array_ref = large_test_array.copy()
    large_test_array_ref += small_test_array

    added_array = add_array(large_test_array, small_test_array, (5, 5))
    assert np.all(added_array == large_test_array_ref)


@pytest.mark.parametrize(
    ("position", "subpixel_index"), zip(test_positions, test_position_indices)
)
def test_subpixel_indices(position, subpixel_index):
    """
    Test subpixel_indices utility function.

    Test by asserting that the function returns correct results for
    given test values.
    """
    assert np.all(subpixel_indices(position, subsampling) == subpixel_index)


class TestCutout2D:
    def setup_class(self):
        self.data = np.arange(20.0).reshape(5, 4)
        self.position = SkyCoord("13h11m29.96s -01d19m18.7s", frame="icrs")
        wcs = WCS(naxis=2)
        rho = np.pi / 3.0
        scale = 0.05 / 3600.0
        wcs.wcs.cd = [
            [scale * np.cos(rho), -scale * np.sin(rho)],
            [scale * np.sin(rho), scale * np.cos(rho)],
        ]
        wcs.wcs.ctype = ["RA---TAN", "DEC--TAN"]
        wcs.wcs.crval = [
            self.position.ra.to_value(u.deg),
            self.position.dec.to_value(u.deg),
        ]
        wcs.wcs.crpix = [3, 3]
        self.wcs = wcs

        # add SIP
        sipwcs = wcs.deepcopy()
        sipwcs.wcs.ctype = ["RA---TAN-SIP", "DEC--TAN-SIP"]
        a = np.array(
            [
                [0, 0, 5.33092692e-08, 3.73753773e-11, -2.02111473e-13],
                [0, 2.44084308e-05, 2.81394789e-11, 5.17856895e-13, 0.0],
                [-2.41334657e-07, 1.29289255e-10, 2.35753629e-14, 0.0, 0.0],
                [-2.37162007e-10, 5.43714947e-13, 0.0, 0.0, 0.0],
                [-2.81029767e-13, 0.0, 0.0, 0.0, 0.0],
            ]
        )
        b = np.array(
            [
                [0, 0, 2.99270374e-05, -2.38136074e-10, 7.23205168e-13],
                [0, -1.71073858e-07, 6.31243431e-11, -5.16744347e-14, 0.0],
                [6.95458963e-06, -3.08278961e-10, -1.75800917e-13, 0.0, 0.0],
                [3.51974159e-11, 5.60993016e-14, 0.0, 0.0, 0.0],
                [-5.92438525e-13, 0.0, 0.0, 0.0, 0.0],
            ]
        )
        sipwcs.sip = Sip(a, b, None, None, wcs.wcs.crpix)
        sipwcs.wcs.set()
        self.sipwcs = sipwcs

    def test_cutout(self):
        sizes = [
            3,
            3 * u.pixel,
            (3, 3),
            (3 * u.pixel, 3 * u.pix),
            (3.0, 3 * u.pixel),
            (2.9, 3.3),
        ]
        for size in sizes:
            position = (2.1, 1.9)
            c = Cutout2D(self.data, position, size)
            assert c.data.shape == (3, 3)
            assert c.data[1, 1] == 10
            assert c.origin_original == (1, 1)
            assert c.origin_cutout == (0, 0)
            assert c.input_position_original == position
            assert_allclose(c.input_position_cutout, (1.1, 0.9))
            assert c.position_original == (2.0, 2.0)
            assert c.position_cutout == (1.0, 1.0)
            assert c.center_original == (2.0, 2.0)
            assert c.center_cutout == (1.0, 1.0)
            assert c.bbox_original == ((1, 3), (1, 3))
            assert c.bbox_cutout == ((0, 2), (0, 2))
            assert c.slices_original == (slice(1, 4), slice(1, 4))
            assert c.slices_cutout == (slice(0, 3), slice(0, 3))

    def test_size_length(self):
        with pytest.raises(ValueError):
            Cutout2D(self.data, (2, 2), (1, 1, 1))

    def test_size_units(self):
        for size in [3 * u.cm, (3, 3 * u.K)]:
            with pytest.raises(ValueError):
                Cutout2D(self.data, (2, 2), size)

    def test_size_pixel(self):
        """
        Check size in derived pixel units.
        """
        size = 0.3 * u.arcsec / (0.1 * u.arcsec / u.pixel)
        c = Cutout2D(self.data, (2, 2), size)
        assert c.data.shape == (3, 3)
        assert c.data[0, 0] == 5
        assert c.slices_original == (slice(1, 4), slice(1, 4))
        assert c.slices_cutout == (slice(0, 3), slice(0, 3))

    def test_size_angle(self):
        c = Cutout2D(self.data, (2, 2), (0.1 * u.arcsec), wcs=self.wcs)
        assert c.data.shape == (2, 2)
        assert c.data[0, 0] == 5
        assert c.slices_original == (slice(1, 3), slice(1, 3))
        assert c.slices_cutout == (slice(0, 2), slice(0, 2))

    def test_size_angle_without_wcs(self):
        with pytest.raises(ValueError):
            Cutout2D(self.data, (2, 2), (3, 3 * u.arcsec))

    def test_cutout_trim_overlap(self):
        c = Cutout2D(self.data, (0, 0), (3, 3), mode="trim")
        assert c.data.shape == (2, 2)
        assert c.data[0, 0] == 0
        assert c.slices_original == (slice(0, 2), slice(0, 2))
        assert c.slices_cutout == (slice(0, 2), slice(0, 2))

    def test_cutout_partial_overlap(self):
        c = Cutout2D(self.data, (0, 0), (3, 3), mode="partial")
        assert c.data.shape == (3, 3)
        assert c.data[1, 1] == 0
        assert c.slices_original == (slice(0, 2), slice(0, 2))
        assert c.slices_cutout == (slice(1, 3), slice(1, 3))

    def test_cutout_partial_overlap_fill_value(self):
        fill_value = -99
        c = Cutout2D(self.data, (0, 0), (3, 3), mode="partial", fill_value=fill_value)
        assert c.data.shape == (3, 3)
        assert c.data[1, 1] == 0
        assert c.data[0, 0] == fill_value

    def test_copy(self):
        data = np.copy(self.data)
        c = Cutout2D(data, (2, 3), (3, 3))
        xy = (0, 0)
        value = 100.0
        c.data[xy] = value
        xy_orig = c.to_original_position(xy)
        yx = xy_orig[::-1]
        assert data[yx] == value

        data = np.copy(self.data)
        c2 = Cutout2D(self.data, (2, 3), (3, 3), copy=True)
        c2.data[xy] = value
        assert data[yx] != value

    def test_to_from_large(self):
        position = (2, 2)
        c = Cutout2D(self.data, position, (3, 3))
        xy = (0, 0)
        result = c.to_cutout_position(c.to_original_position(xy))
        assert_allclose(result, xy)

    def test_skycoord_without_wcs(self):
        with pytest.raises(ValueError):
            Cutout2D(self.data, self.position, (3, 3))

    def test_skycoord(self):
        c = Cutout2D(self.data, self.position, (3, 3), wcs=self.wcs)
        skycoord_original = self.position.from_pixel(
            c.center_original[1], c.center_original[0], self.wcs
        )
        skycoord_cutout = self.position.from_pixel(
            c.center_cutout[1], c.center_cutout[0], c.wcs
        )
        assert_quantity_allclose(skycoord_original.ra, skycoord_cutout.ra)
        assert_quantity_allclose(skycoord_original.dec, skycoord_cutout.dec)

    def test_skycoord_partial(self):
        c = Cutout2D(self.data, self.position, (3, 3), wcs=self.wcs, mode="partial")
        skycoord_original = self.position.from_pixel(
            c.center_original[1], c.center_original[0], self.wcs
        )
        skycoord_cutout = self.position.from_pixel(
            c.center_cutout[1], c.center_cutout[0], c.wcs
        )
        assert_quantity_allclose(skycoord_original.ra, skycoord_cutout.ra)
        assert_quantity_allclose(skycoord_original.dec, skycoord_cutout.dec)

    def test_naxis_update(self):
        xsize = 2
        ysize = 3
        c = Cutout2D(self.data, self.position, (ysize, xsize), wcs=self.wcs)
        assert c.wcs.array_shape == (ysize, xsize)

    def test_crpix_maps_to_crval(self):
        w = Cutout2D(self.data, (0, 0), (3, 3), wcs=self.sipwcs, mode="partial").wcs
        pscale = np.sqrt(proj_plane_pixel_area(w))
        assert_allclose(
            w.wcs_pix2world(*w.wcs.crpix, 1), w.wcs.crval, rtol=0.0, atol=1e-6 * pscale
        )
        assert_allclose(
            w.all_pix2world(*w.wcs.crpix, 1), w.wcs.crval, rtol=0.0, atol=1e-6 * pscale
        )

    def test_cutout_with_nddata_as_input(self):
        # This is essentially a copy/paste of test_skycoord with the
        # input a ccd with wcs attribute instead of passing the
        # wcs separately.
        ccd = CCDData(data=self.data, wcs=self.wcs, unit="adu")
        c = Cutout2D(ccd, self.position, (3, 3))
        skycoord_original = self.position.from_pixel(
            c.center_original[1], c.center_original[0], self.wcs
        )
        skycoord_cutout = self.position.from_pixel(
            c.center_cutout[1], c.center_cutout[0], c.wcs
        )
        assert_quantity_allclose(skycoord_original.ra, skycoord_cutout.ra)
        assert_quantity_allclose(skycoord_original.dec, skycoord_cutout.dec)


def test_cutout_section(tmp_path):
    # Make sure that one can pass ImageHDU.section and CompImageHDU.section
    # to Cutout2D

    data = np.ones((200, 200))

    hdu = fits.ImageHDU(data=data)
    hdu.writeto(tmp_path / "uncompressed.fits")

    with fits.open(tmp_path / "uncompressed.fits") as hdul:
        c = Cutout2D(
            hdul[1].section,
            (75, 75),
            100 * u.pix,
        )

    chdu = fits.CompImageHDU(data=data)
    chdu.writeto(tmp_path / "compressed.fits")

    with fits.open(tmp_path / "compressed.fits") as hdul:
        c = Cutout2D(
            hdul[1].section,
            (75, 75),
            100 * u.pix,
        )
